/*****************************************************************************/
/*
                                 reCAPTCHA.c

Google reCAPTCHA.

http://en.wikipedia.org/wiki/ReCAPTCHA
http://www.google.com/recaptcha
http://code.google.com/apis/recaptcha/docs/verify.html

  A CAPTCHA is a program that can tell whether its user is a human or a
  computer. You've probably seen them - colorful images with distorted text
  at the bottom of Web registration forms. CAPTCHAs are used by many websites
  to prevent abuse from "bots," or automated programs usually written to
  generate spam. No computer program can read distorted text as well as humans
  can, so bots cannot navigate sites protected by CAPTCHAs.

Google Mailhide.

http://www.google.com/recaptcha/mailhide/
http://code.google.com/apis/recaptcha/docs/mailhideapi.html

All the credit for this excellent idea goes to Tim Sneddon.  $$TICKET$$ is the
original (and somewhat) lame attempt to provision against the same problem. 
reCAPTCHA is a much more robust (if externally dependent) solution.  Insomnia
(and familiarity with the code) meant that my implementation got to the
test-bench first.  Thanks Tim!


VERSION HISTORY
---------------
18-JUL-2011  MGD  initial
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <iodef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <starlet.h>

/* Internet-related header files */
#include <socket.h>
#include <in.h>
#include <netdb.h>
#include <inet.h>

#include "vwcms.h"
#include "aes.h"
#include "recaptcha.h"
#include "cgilib.h"
#include "content.h"
#include "config.h"
#include "other.h"
#include "template.h"

/* define required values from UCX$INETDEF.H (Multinet does not supply one) */
#define INET_PROTYP$C_STREAM 1
#define INETACP$C_TRANS 2
#define INETACP_FUNC$C_GETHOSTBYNAME 1
#define INETACP_FUNC$C_GETHOSTBYADDR 2
#define UCX$C_AF_INET 2
#define UCX$C_DSC_ALL 2
#define UCX$C_FULL_DUPLEX_CLOSE 8192
#define UCX$C_REUSEADDR 4
#define UCX$C_SOCK_NAME 4
#define UCX$C_SOCKOPT 1
#define UCX$C_TCP 6

#define RMS$_RNF 98994

#define FI_LI "RECAPTCHA", __LINE__

/******************/
/* global storage */
/******************/

BOOL  ReCaptchaAfter,
      ReCaptchaBefore;

static struct AnIOsb {
   unsigned short  iosb$w_status;
   unsigned short  iosb$w_bcnt;
   unsigned int    iosb$l_reserved;
};

static $DESCRIPTOR (InetDeviceDsc, "UCX$DEVICE");

static unsigned short  ReCaptchaChan;

static int  OptionEnabled = 1;

static struct {
   unsigned short  Length;
   unsigned short  Parameter;
   void  *Address;
} ReuseAddress =
   { sizeof(OptionEnabled), UCX$C_REUSEADDR, &OptionEnabled },
  ReuseAddressSocketOption =
   { sizeof(ReuseAddress), UCX$C_SOCKOPT, &ReuseAddress };

static unsigned short  TcpSocket [3] =
   { UCX$C_TCP, INET_PROTYP$C_STREAM, UCX$C_AF_INET };

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern char  *CgiRemoteAddrPtr;

extern CONTENT_DATA  *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/****************************************************************************/
/*
Implement the $$RECAPTCHA$$ directive.
*/

void ReCaptchaDirective (char *SubsArg)

{
   char  *cptr;
   char  DisplayAs [256] = "",
         EmailAddress [256] = "",
         RevealPrompt [256] = "";

   /*********/
   /* begin */
   /*********/

   if (!SubsArg)
   {
      /* initialise */
      ReCaptchaAfter = ReCaptchaBefore = FALSE;
      return;
   }

   if (*(cptr = SubsArg) == '$') cptr++;
   /* the default in the absence of any parameter */
   if (!*cptr) ReCaptchaBefore = TRUE;
   while (*cptr)
   {
      if (strsame (cptr, "display=", 8))
         cptr = TemplateParseParam (cptr+8, DisplayAs, sizeof(DisplayAs));
      else
      if (strsame (cptr, "reveal=", 7))
         cptr = TemplateParseParam (cptr+7, RevealPrompt, sizeof(RevealPrompt));
      else
      if (strsame (cptr, "mailhide=", 9))
         cptr = TemplateParseParam (cptr+9, EmailAddress, sizeof(EmailAddress));
      else
      if (strsame (cptr, "after=submit", 12))
      {
         ReCaptchaAfter = TRUE;
         ReCaptchaBefore = FALSE;
         cptr = TemplateParseSkipParam (cptr+12, NULL);
      }
      else
      if (strsame (cptr, "before=submit", 13))
      {
         ReCaptchaAfter = FALSE;
         ReCaptchaBefore = TRUE;
         cptr = TemplateParseSkipParam (cptr+13, NULL);
      }
      else
      if (strsame (cptr, "submit", 6))
      {
         ReCaptchaAfter = FALSE;
         ReCaptchaBefore = TRUE;
         cptr = TemplateParseSkipParam (cptr+6, NULL);
      }
      else
      {
         TemplateParamError ("RECAPTCHA", SubsArg, cptr);
         return;
      }
   }

   if (EmailAddress[0])
      ReCaptchaMailHide (EmailAddress, DisplayAs, RevealPrompt);
}

/****************************************************************************/
/*
Include the Google reCAPTCHA HTML/JavaScript code on the page. 
*/

void ReCaptchaChallenge ()

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (!VWcmsConfig.ReCaptchaPrivateKey)
   {
      TemplateAdvise ("EreCAPTCHA private key not configured");
      return;
   }
   if (!VWcmsConfig.ReCaptchaPublicKey)
   {
      TemplateAdvise ("EreCAPTCHA public key not configured");
      return;
   }

   /* try to second-guess to a avoid a mix of scheme elements on the page */
   if (!strcmp (GetCgiVar("REQUEST_SCHEME"), "https:") ||
       !strcmp (GetCgiVar("SERVER_PORT"), "443"))
      cptr = "https:";
   else
      cptr = "http:";

   fprintf (stdout,
"<p><script type=\"text/javascript\" \
src=\"%s//www.google.com/recaptcha/api/challenge?k=%s\">\
</script>\
<noscript>\
<iframe src=\"%s//www.google.com/recaptcha/api/noscript?k=%s\" \
height=\"300\" width=\"500\" frameborder=\"0\"></iframe><br />\
<textarea name=\"recaptcha_challenge_field\" rows=\"3\" cols=\"40\">\
</textarea>\
<input type=\"hidden\" name=\"recaptcha_response_field\" \
value=\"manual_challenge\" />\
</noscript></p>\n",
            cptr, VWcmsConfig.ReCaptchaPublicKey,
            cptr, VWcmsConfig.ReCaptchaPublicKey);
}

/****************************************************************************/
/*
Hard-wired Google reCAPTCHA verification API request.
Returns NULL if verified or a string indicating the error.
*/

char* ReCaptchaVerify ()

{
   static  $DESCRIPTOR (BodyFaoDsc,
"privatekey=!AZ&remoteip=!AZ&challenge=!AZ&response=!AZ\0");

   static  $DESCRIPTOR (RequestFaoDsc,
"POST /recaptcha/api/verify HTTP/1.0\r\n\
Host: !AZ\r\n\
Content-Type: application/x-www-form-urlencoded\r\n\
Content-Length: !UL\r\n\
Connection: close\r\n\
\r\n\
!AZ\0");

   int  status,
        IpAddress;
   unsigned short  BodyLength,
                   RemotePort = 80,
                   RequestLength;
   unsigned long  OneSecondDelta [2] = { -10000000, -1 };
   char  *cptr, *sptr, *zptr,
         *ChallengeField,
         *ResponseField,
         *RemoteName = "www.google.com";
   char  HttpRequest [1024],
         RequestBody [512],
         ResponseBuffer [1024];
   struct AnIOsb  IOsb;
   struct sockaddr_in  SocketName;
   struct hostent  *HostEntryPtr;
   $DESCRIPTOR (HttpRequestDsc, HttpRequest);
   $DESCRIPTOR (RequestBodyDsc, RequestBody);

   struct {
      unsigned long  Length;
      void  *Address;
      int  *LengthPtr;
   } SocketNameItem =
      { sizeof(SocketName), &SocketName, 0 };

   /*********/
   /* begin */
   /*********/

   if (!VWcmsConfig.ReCaptchaPublicKey || !VWcmsConfig.ReCaptchaPrivateKey)
      return ("bugcheck!");

   ChallengeField = GetCgiVarNull ("FORM_RECAPTCHA_CHALLENGE_FIELD");
   ResponseField = GetCgiVarNull ("FORM_RECAPTCHA_RESPONSE_FIELD");
   if (!ChallengeField || !ChallengeField[0] || !ResponseField)
      return ("bugcheck!");

   if (!ResponseField[0]) return (CONFIG_LANG_PARSE(LangReCaptcha,1));

   if (WatchEnabled)
      WatchThis (FI_LI, "TCPIP connect !AZ:!UL", RemoteName, RemotePort);

   if (!(HostEntryPtr = gethostbyname (RemoteName)))
   {
      status = vaxc$errno;
      if (status == RMS$_RNF) status = SS$_NOSUCHNODE;
      if (status == SS$_ENDOFFILE) status = SS$_NOSUCHNODE;
      goto ReCaptchaVerifyNotReachable;
   }
   memcpy (&IpAddress, HostEntryPtr->h_addr, 4);

   /***********/
   /* connect */
   /***********/

   /* assign a channel to the internet template device */
   if (VMSnok (status = sys$assign (&InetDeviceDsc, &ReCaptchaChan, 0, 0)))
      return ("bugcheck!");

   /* make the channel a TCP, connection-oriented socket */
   status = sys$qiow (0, ReCaptchaChan, IO$_SETMODE, &IOsb, 0, 0,
                      &TcpSocket, 0, 0, 0, 0, 0);
   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      sys$dassgn (ReCaptchaChan);
      ReCaptchaChan = 0;
      return ("bugcheck!");
   }

   SocketName.sin_family = AF_INET;
   SocketName.sin_port = htons (RemotePort);
   memcpy (&SocketName.sin_addr.s_addr, &IpAddress, 4);

   /* let's be patient but only for so long */
   OneSecondDelta[0] *= 15;
   status = sys$setimr (0, &OneSecondDelta, ReCaptchaTimeout,
                                            ReCaptchaTimeout, 0);
   if (VMSnok (status))
   {
      sys$dassgn (ReCaptchaChan);
      ReCaptchaChan = 0;
      return ("bugcheck!");
   }

   status = sys$qiow (0, ReCaptchaChan, IO$_ACCESS, &IOsb, 0, 0,
                      0, 0, &SocketNameItem, 0, 0, 0);
   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) goto ReCaptchaVerifyNotReachable;

   /****************/
   /* POST request */
   /****************/

   status = sys$fao (&BodyFaoDsc, &BodyLength, &RequestBodyDsc,
                     CgiLibUrlEncode(VWcmsConfig.ReCaptchaPrivateKey,-1,NULL,0),
                     CgiRemoteAddrPtr,
                     CgiLibUrlEncode(ChallengeField,-1,NULL,0),
                     CgiLibUrlEncode(ResponseField,-1,NULL,0));
   if (VMSnok (status) || status == SS$_BUFFEROVF) ErrorExit (status, FI_LI);
   BodyLength--;

   status = sys$fao (&RequestFaoDsc, &RequestLength, &HttpRequestDsc,
                     RemoteName,
                     BodyLength,
                     RequestBody);
   if (VMSnok (status) || status == SS$_BUFFEROVF) ErrorExit (status, FI_LI);
   RequestLength--;

   if (WatchEnabled) WatchThis (FI_LI, "REQUEST !AZ", HttpRequest);

   status = sys$qiow (0, ReCaptchaChan, IO$_WRITEVBLK, &IOsb, 0, 0,
                      HttpRequest, RequestLength, 0, 0, 0, 0);
   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) goto ReCaptchaVerifyNotReachable;

   /*****************/
   /* read response */
   /*****************/

   status = sys$qiow (0, ReCaptchaChan, IO$_READVBLK, &IOsb, 0, 0,
                      ResponseBuffer, sizeof(ResponseBuffer)-1, 0, 0, 0, 0);
   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status)) goto ReCaptchaVerifyNotReachable;

   /* and close the connection */
   sys$cantim (ReCaptchaTimeout, 0);
   sys$qiow (0, ReCaptchaChan, IO$_DEACCESS, 0, 0, 0,
             0, 0, 0, 0, 0, 0);
   sys$dassgn (ReCaptchaChan);
   ReCaptchaChan = 0;

   ResponseBuffer[IOsb.iosb$w_bcnt] = '\0';
   if (WatchEnabled) WatchThis (FI_LI, "RESPONSE !AZ", ResponseBuffer);

   /********************/
   /* process response */
   /********************/

   /* locate the response body */
   for (cptr = ResponseBuffer; *cptr; cptr++)
   {
      if (*(USHORTPTR)cptr == '\n\n') { cptr += 2; break; }
      if (*(ULONGPTR)cptr == '\r\n\r\n') { cptr += 4; break; }
   }

   if (!strncmp (cptr, "true\n", 5)) return (NULL);

   /* interpret the error string */
   while (*cptr && *cptr != '\n') cptr++;
   if (*cptr) cptr++;
   for (sptr = cptr; *sptr && *sptr != '\n'; sptr++);
   *sptr = '\0';

   /* munge this common error into something a bit more friendly */
   if (!strcmp (cptr, "incorrect-captcha-sol"))
      cptr = CONFIG_LANG_PARSE(LangReCaptcha,3);

   return (cptr);

ReCaptchaVerifyNotReachable:

   sys$cantim (ReCaptchaTimeout, 0);
   sys$dassgn (ReCaptchaChan);
   ReCaptchaChan = 0;
   cptr = CgiLibVeeMemCalloc(256);
   sprintf (cptr, "%s (%%X%08.08X)",
            CONFIG_LANG_PARSE(LangReCaptcha,2), status);
   return (cptr);
}

/****************************************************************************/
/*
Simply break the connection.
*/

void  ReCaptchaTimeout ()

{
   /*********/
   /* begin */
   /*********/

   sys$cancel (ReCaptchaChan);
   sys$dassgn (ReCaptchaChan);
   ReCaptchaChan = 0;
}

/****************************************************************************/
/*
Include the Google Mailhide HTML/JavaScript code on the page. 
*/

void ReCaptchaMailHide
(
char *EmailAddress,
char *DisplayAs,
char *RevealPrompt
)
{
   static char  ErrorAddress [] = "<a href="">...@error</a>";

   int  cnt, len;
   char  ch;
   char  *cptr, *sptr, *zptr, *b64;
   char  LocalDisplayAs [256];
   unsigned char  inbuf [256],
                  invec [16],
                  outbuf [256],
                  prkey16 [16],
                  tmp [3] = { 0,0,0 };
   aes_context  aesctx;

   /*********/
   /* begin */
   /*********/

   if (!VWcmsConfig.ReCaptchaMailHidePrivateKey)
   {
      fputs (ErrorAddress, stdout);
      TemplateAdvise ("EMAILHIDE private key not configured");
      return;
   }
   if (!VWcmsConfig.ReCaptchaMailHidePublicKey)
   {
      fputs (ErrorAddress, stdout);
      TemplateAdvise ("EMAILHIDE public key not configured");
      return;
   }

   if (!DisplayAs || !DisplayAs[0])
   {
      zptr = (sptr = LocalDisplayAs) + sizeof(LocalDisplayAs)-1;
      for (cptr = "..."; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = EmailAddress; *cptr; cptr++);
      while (cptr > EmailAddress && *cptr != '@') cptr--;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      DisplayAs = LocalDisplayAs;
   }

   if (!RevealPrompt || !RevealPrompt[0])
      RevealPrompt = "Reveal this e-mail address";

   /* convert hex string to octets (klunky perhaps but it works) */
   cptr = VWcmsConfig.ReCaptchaMailHidePrivateKey;
   for (cnt = 0; cnt < 16; cnt++)
   {
      if (!(tmp[0] = *cptr++)) break;
      if (!(tmp[1] = *cptr++)) break;
      if (!((tmp[0] >= '0' && tmp[0] <= '9') ||
            (tmp[0] >= 'a' && tmp[0] <= 'f'))) break;
      if (!((tmp[1] >= '0' && tmp[1] <= '9') ||
            (tmp[1] >= 'a' && tmp[1] <= 'f'))) break;
      prkey16[cnt] = strtoul((char*)tmp,0,16); 
   }
   if (cnt != 16)
   {
      fputs (ErrorAddress, stdout);
      TemplateAdvise ("EMAILHIDE private key unusable");
      return;
   }

   /* copy mail address, padding to 16 byte boundary as required by CBC */
   zptr = (sptr = (char*)inbuf) + sizeof(inbuf);
   cnt = 0;
   for (cptr = EmailAddress; *cptr && sptr < zptr; *sptr++ = *cptr++) cnt++;
   ch = 16 - (cnt % 16);
   while ((cnt++ % 16) && sptr < zptr) *sptr++ = ch;
   len = sptr - (char*)inbuf;
   memset (outbuf, 0, sizeof(outbuf));

   memset (invec, 0, sizeof(invec));
   if (aes_setkey_enc (&aesctx, prkey16, 128))
   {
      fputs (ErrorAddress, stdout);
      TemplateAdvise ("EMAILHIDE private key rejected");
      return;
   }
   if (aes_crypt_cbc (&aesctx, AES_ENCRYPT, len, invec, inbuf, outbuf))
   {
      fputs (ErrorAddress, stdout);
      TemplateAdvise ("EMAILHIDE encrypt failed");
      return;
   }

   b64 = Base64UrlSafeEncode ((char*)outbuf, len);

   /* try to second-guess to a avoid a mix of scheme elements on the page */
   if (!strcmp (GetCgiVar("REQUEST_SCHEME"), "https:") ||
       !strcmp (GetCgiVar("SERVER_PORT"), "443"))
      cptr = "https:";
   else
      cptr = "http:";

   fprintf (stdout,
"<a href=\"%s//www.google.com/recaptcha/mailhide/d?k=%s&c=%s\" \
onclick=\"window.open(\'%s//www.google.com/recaptcha/mailhide/d?k=%s&c=%s\',\
\'\',\'toolbar=0,scrollbars=0,location=0,statusbar=0,menubar=0,resizable=0,\
width=500,height=300\');return false;\" title=\"%s\">%s</a>",
            cptr, VWcmsConfig.ReCaptchaMailHidePublicKey, b64,
            cptr, VWcmsConfig.ReCaptchaMailHidePublicKey, b64,
            HtmlEscape(RevealPrompt),
            HtmlEscape(DisplayAs));
}

/*****************************************************************************/
/*
Returns a pointer if the section contains a $$RECAPTCHA$$ directive that does
not contain a "mailhide=" parameter (i.e. a reCAPTCHA challenge), else NULL.
*/

char* ReCaptchaPresent ()

{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReCaptchaPresent()\n");

   for (cptr = ContentSectionPtr->TextPtr; *cptr;)
   {
      if (!*(cptr = TemplateParseFind (cptr, "$$RECAPTCHA$"))) break;
      sptr = TemplateParseFindParam (cptr+12, "mailhide=");
      if (*(USHORTPTR)sptr == '$$') return (cptr);
   }
   return (NULL);
}

/****************************************************************************/

